/*
 * livepatch_bsc1196959
 *
 * Fix for CVE-2021-39698, bsc#1196959
 *
 *  Upstream commits:
 *  42288cb44c4b ("wait: add wake_up_pollfree()")
 *  a880b28a71e3 ("binder: use wake_up_pollfree()")
 *  9537bae0da1f ("signalfd: use wake_up_pollfree()"
 *  363bee27e258 ("aio: keep poll requests on waitqueue until completed")
 *  50252e4b5e98 ("aio: fix use-after-free due to missing POLLFREE handling")
 *  791f3465c4af ("io_uring: fix UAF due to missing POLLFREE handling")
 *
 *  SLE12-SP3 commit:
 *  not affected
 *
 *  SLE12-SP4, SLE12-SP5, SLE15 and SLE15-SP1 commit:
 *  not affected
 *
 *  SLE15-SP2 and -SP3 commits:
 *  fbcad1232a2939a59b964cee9145dbaf4fa1e78c
 *  00c1c8d55a7a4fdedb36c7993cb09c9814a4742f
 *  28c3691b228694917e6bcd8822fc16c97c7ffac7
 *  3ae391179ed2a4ee434cfbc4e932906dd8fbe8fa
 *  b026506dc6444c526bb88c0c33100b5843e60d0b
 *
 *  SLE15-SP4 commits:
 *  ed751ab39509ea108c8aeb3dcf376952d1add35d
 *  ea70040e7d1cea83acca136adea869c75216cb44
 *  799ed47b18c245705b9b938e6062b92dc83eb00c
 *  8a0814bb6f78da58ce1ded5d1a3fd4ddc22c75df
 *  ee17f5cf5e18b0109d704e031cc08428c15db650
 *
 *
 *  Copyright (c) 2022 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* klp-ccp: from fs/io_uring.c */
#include <linux/kernel.h>
#include <linux/init.h>
#include <linux/errno.h>
#include <linux/syscalls.h>

/* klp-ccp: from include/linux/fs.h */
static int (*klpe_sync_file_range)(struct file *file, loff_t offset, loff_t nbytes,
				unsigned int flags);

/* klp-ccp: from include/linux/mm.h */
static int (*klpe_do_madvise)(struct mm_struct *mm, unsigned long start, size_t len_in, int behavior);

/* klp-ccp: from include/linux/socket.h */
static int (*klpe___sys_accept4_file)(struct file *file, unsigned file_flags,
			struct sockaddr __user *upeer_sockaddr,
			 int __user *upeer_addrlen, int flags,
			 unsigned long nofile);

static int (*klpe___sys_shutdown_sock)(struct socket *sock, int how);

/* klp-ccp: from fs/io_uring.c */
#include <linux/compat.h>
#include <linux/refcount.h>
#include <linux/uio.h>
#include <linux/bits.h>
#include <linux/sched/signal.h>
#include <linux/fs.h>
#include <linux/file.h>
#include <linux/fdtable.h>
#include <linux/mm.h>
#include <linux/percpu.h>
#include <linux/slab.h>
#include <linux/blkdev.h>
#include <linux/bvec.h>
#include <linux/net.h>

/* klp-ccp: from include/linux/splice.h */
static long (*klpe_do_splice)(struct file *in, loff_t *off_in,
		      struct file *out, loff_t *off_out,
		      size_t len, unsigned int flags);

static long (*klpe_do_tee)(struct file *in, struct file *out, size_t len,
		   unsigned int flags);

/* klp-ccp: from fs/io_uring.c */
#include <net/scm.h>
#include <linux/sched/mm.h>
#include <linux/uaccess.h>
#include <linux/nospec.h>
#include <linux/sizes.h>

#include <linux/highmem.h>
#include <linux/fadvise.h>

/* klp-ccp: from include/linux/eventpoll.h */
#ifdef CONFIG_EPOLL

static int (*klpe_do_epoll_ctl)(int epfd, int op, int fd, struct epoll_event *epds,
		 bool nonblock);

#else
#error "klp-ccp: non-taken branch"
#endif

/* klp-ccp: from fs/io_uring.c */
#include <linux/splice.h>
#include <linux/task_work.h>
#include <linux/pagemap.h>
#include <uapi/linux/io_uring.h>

/* klp-ccp: from fs/internal.h */
static long (*klpe_do_rmdir)(int dfd, struct filename *name);
static long (*klpe_do_unlinkat)(int dfd, struct filename *name);

static int (*klpe_do_renameat2)(int olddfd, struct filename *oldname, int newdfd,
		 struct filename *newname, unsigned int flags);

static int (*klpe___close_fd_get_file)(unsigned int fd, struct file **res);

static int (*klpe_do_statx)(int dfd, const char __user *filename, unsigned flags,
	     unsigned int mask, struct statx __user *buffer);

/* klp-ccp: from fs/io-wq.h */
#include <linux/refcount.h>

struct io_wq_work_node {
	struct io_wq_work_node *next;
};

struct io_wq_work {
	struct io_wq_work_node list;
	unsigned flags;
};

static inline bool io_wq_current_is_worker(void)
{
	return in_task() && (current->flags & PF_IO_WORKER) &&
		current->pf_io_worker;
}

/* klp-ccp: from fs/io_uring.c */
#define IO_REQ_CLEAN_FLAGS (REQ_F_BUFFER_SELECTED | REQ_F_NEED_CLEANUP | \
				REQ_F_POLLED | REQ_F_INFLIGHT | REQ_F_CREDS)

struct io_uring {
	u32 head ____cacheline_aligned_in_smp;
	u32 tail ____cacheline_aligned_in_smp;
};

struct io_rings {
	/*
	 * Head and tail offsets into the ring; the offsets need to be
	 * masked to get valid indices.
	 *
	 * The kernel controls head of the sq ring and the tail of the cq ring,
	 * and the application controls tail of the sq ring and the head of the
	 * cq ring.
	 */
	struct io_uring		sq, cq;
	/*
	 * Bitmasks to apply to head and tail offsets (constant, equals
	 * ring_entries - 1)
	 */
	u32			sq_ring_mask, cq_ring_mask;
	/* Ring sizes (constant, power of 2) */
	u32			sq_ring_entries, cq_ring_entries;
	/*
	 * Number of invalid entries dropped by the kernel due to
	 * invalid index stored in array
	 *
	 * Written by the kernel, shouldn't be modified by the
	 * application (i.e. get number of "new events" by comparing to
	 * cached value).
	 *
	 * After a new SQ head value was read by the application this
	 * counter includes all submissions that were dropped reaching
	 * the new SQ head (and possibly more).
	 */
	u32			sq_dropped;
	/*
	 * Runtime SQ flags
	 *
	 * Written by the kernel, shouldn't be modified by the
	 * application.
	 *
	 * The application needs a full memory barrier before checking
	 * for IORING_SQ_NEED_WAKEUP after updating the sq tail.
	 */
	u32			sq_flags;
	/*
	 * Runtime CQ flags
	 *
	 * Written by the application, shouldn't be modified by the
	 * kernel.
	 */
	u32			cq_flags;
	/*
	 * Number of completion events lost because the queue was full;
	 * this should be avoided by the application by making sure
	 * there are not more requests pending than there is space in
	 * the completion queue.
	 *
	 * Written by the kernel, shouldn't be modified by the
	 * application (i.e. get number of "new events" by comparing to
	 * cached value).
	 *
	 * As completion events come in out of order this counter is not
	 * ordered with any other data.
	 */
	u32			cq_overflow;
	/*
	 * Ring buffer of completion events.
	 *
	 * The kernel writes completion events fresh every time they are
	 * produced, so the application is allowed to modify pending
	 * entries.
	 */
	struct io_uring_cqe	cqes[] ____cacheline_aligned_in_smp;
};

enum io_uring_cmd_flags {
	IO_URING_F_NONBLOCK		= 1,
	IO_URING_F_COMPLETE_DEFER	= 2,
};

struct io_file_table {
	struct io_fixed_file *files;
};

struct io_buffer {
	struct list_head list;
	__u64 addr;
	__u32 len;
	__u16 bid;
};

struct io_restriction {
	DECLARE_BITMAP(register_op, IORING_REGISTER_LAST);
	DECLARE_BITMAP(sqe_op, IORING_OP_LAST);
	u8 sqe_flags_allowed;
	u8 sqe_flags_required;
	bool registered;
};

struct io_sq_data {
	refcount_t		refs;
	atomic_t		park_pending;
	struct mutex		lock;

	/* ctx's that are using this sqd */
	struct list_head	ctx_list;

	struct task_struct	*thread;
	struct wait_queue_head	wait;

	unsigned		sq_thread_idle;
	int			sq_cpu;
	pid_t			task_pid;
	pid_t			task_tgid;

	unsigned long		state;
	struct completion	exited;
};

#define IO_COMPL_BATCH			32
#define IO_REQ_CACHE_SIZE		32

struct io_submit_link {
	struct io_kiocb		*head;
	struct io_kiocb		*last;
};

struct io_submit_state {
	struct blk_plug		plug;
	struct io_submit_link	link;

	/*
	 * io_kiocb alloc cache
	 */
	void			*reqs[IO_REQ_CACHE_SIZE];
	unsigned int		free_reqs;

	bool			plug_started;

	/*
	 * Batch completion logic
	 */
	struct io_kiocb		*compl_reqs[IO_COMPL_BATCH];
	unsigned int		compl_nr;
	/* inline/task_work completion list, under ->uring_lock */
	struct list_head	free_list;

	unsigned int		ios_left;
};

struct io_ring_ctx {
	/* const or read-mostly hot data */
	struct {
		struct percpu_ref	refs;

		struct io_rings		*rings;
		unsigned int		flags;
		unsigned int		compat: 1;
		unsigned int		drain_next: 1;
		unsigned int		eventfd_async: 1;
		unsigned int		restricted: 1;
		unsigned int		off_timeout_used: 1;
		unsigned int		drain_active: 1;
	} ____cacheline_aligned_in_smp;

	/* submission data */
	struct {
		struct mutex		uring_lock;

		/*
		 * Ring buffer of indices into array of io_uring_sqe, which is
		 * mmapped by the application using the IORING_OFF_SQES offset.
		 *
		 * This indirection could e.g. be used to assign fixed
		 * io_uring_sqe entries to operations and only submit them to
		 * the queue when needed.
		 *
		 * The kernel modifies neither the indices array nor the entries
		 * array.
		 */
		u32			*sq_array;
		struct io_uring_sqe	*sq_sqes;
		unsigned		cached_sq_head;
		unsigned		sq_entries;
		struct list_head	defer_list;

		/*
		 * Fixed resources fast path, should be accessed only under
		 * uring_lock, and updated through io_uring_register(2)
		 */
		struct io_rsrc_node	*rsrc_node;
		struct io_file_table	file_table;
		unsigned		nr_user_files;
		unsigned		nr_user_bufs;
		struct io_mapped_ubuf	**user_bufs;

		struct io_submit_state	submit_state;
		struct list_head	timeout_list;
		struct list_head	cq_overflow_list;
		struct xarray		io_buffers;
		struct xarray		personalities;
		u32			pers_next;
		unsigned		sq_thread_idle;
	} ____cacheline_aligned_in_smp;

	/* IRQ completion list, under ->completion_lock */
	struct list_head	locked_free_list;
	unsigned int		locked_free_nr;

	const struct cred	*sq_creds;	/* cred used for __io_sq_thread() */
	struct io_sq_data	*sq_data;	/* if using sq thread polling */

	struct wait_queue_head	sqo_sq_wait;
	struct list_head	sqd_list;

	unsigned long		check_cq_overflow;

	struct {
		unsigned		cached_cq_tail;
		unsigned		cq_entries;
		struct eventfd_ctx	*cq_ev_fd;
		struct wait_queue_head	poll_wait;
		struct wait_queue_head	cq_wait;
		unsigned		cq_extra;
		atomic_t		cq_timeouts;
		unsigned		cq_last_tm_flush;
	} ____cacheline_aligned_in_smp;

	struct {
		spinlock_t		completion_lock;

		spinlock_t		timeout_lock;

		/*
		 * ->iopoll_list is protected by the ctx->uring_lock for
		 * io_uring instances that don't use IORING_SETUP_SQPOLL.
		 * For SQPOLL, only the single threaded io_sq_thread() will
		 * manipulate the list, hence no extra locking is needed there.
		 */
		struct list_head	iopoll_list;
		struct hlist_head	*cancel_hash;
		unsigned		cancel_hash_bits;
		bool			poll_multi_queue;
	} ____cacheline_aligned_in_smp;

	struct io_restriction		restrictions;

	/* slow path rsrc auxilary data, used by update/register */
	struct {
		struct io_rsrc_node		*rsrc_backup_node;
		struct io_mapped_ubuf		*dummy_ubuf;
		struct io_rsrc_data		*file_data;
		struct io_rsrc_data		*buf_data;

		struct delayed_work		rsrc_put_work;
		struct llist_head		rsrc_put_llist;
		struct list_head		rsrc_ref_list;
		spinlock_t			rsrc_ref_lock;
	};

	/* Keep this last, we don't need it for the fast path */
	struct {
#if defined(CONFIG_UNIX)
		struct socket		*ring_sock;
#else
#error "klp-ccp: a preceeding branch should have been taken"
#endif
		struct io_wq_hash		*hash_map;

		/* Only used for accounting purposes */
		struct user_struct		*user;
		struct mm_struct		*mm_account;

		/* ctx exit and cancelation */
		struct llist_head		fallback_llist;
		struct delayed_work		fallback_work;
		struct work_struct		exit_work;
		struct list_head		tctx_list;
		struct completion		ref_comp;
	};
};

struct io_poll_iocb {
	struct file			*file;
	struct wait_queue_head		*head;
	__poll_t			events;
	bool				done;
	bool				canceled;
	struct wait_queue_entry		wait;
};

struct io_poll_update {
	struct file			*file;
	u64				old_user_data;
	u64				new_user_data;
	__poll_t			events;
	bool				update_events;
	bool				update_user_data;
};

struct io_close {
	struct file			*file;
	int				fd;
};

struct io_timeout_data {
	struct io_kiocb			*req;
	struct hrtimer			timer;
	struct timespec64		ts;
	enum hrtimer_mode		mode;
};

struct io_accept {
	struct file			*file;
	struct sockaddr __user		*addr;
	int __user			*addr_len;
	int				flags;
	unsigned long			nofile;
};

struct io_sync {
	struct file			*file;
	loff_t				len;
	loff_t				off;
	int				flags;
	int				mode;
};

struct io_cancel {
	struct file			*file;
	u64				addr;
};

struct io_timeout {
	struct file			*file;
	u32				off;
	u32				target_seq;
	struct list_head		list;
	/* head of the link, used by linked timeouts only */
	struct io_kiocb			*head;
	/* for linked completions */
	struct io_kiocb			*prev;
};

struct io_timeout_rem {
	struct file			*file;
	u64				addr;

	/* timeout update */
	struct timespec64		ts;
	u32				flags;
};

struct io_rw {
	/* NOTE: kiocb has the file as the first member, so don't do it here */
	struct kiocb			kiocb;
	u64				addr;
	u64				len;
};

struct io_connect {
	struct file			*file;
	struct sockaddr __user		*addr;
	int				addr_len;
};

struct io_sr_msg {
	struct file			*file;
	union {
		struct compat_msghdr __user	*umsg_compat;
		struct user_msghdr __user	*umsg;
		void __user			*buf;
	};
	int				msg_flags;
	int				bgid;
	size_t				len;
	struct io_buffer		*kbuf;
};

struct io_open {
	struct file			*file;
	int				dfd;
	struct filename			*filename;
	struct open_how			how;
	unsigned long			nofile;
};

struct io_rsrc_update {
	struct file			*file;
	u64				arg;
	u32				nr_args;
	u32				offset;
};

struct io_fadvise {
	struct file			*file;
	u64				offset;
	u32				len;
	u32				advice;
};

struct io_madvise {
	struct file			*file;
	u64				addr;
	u32				len;
	u32				advice;
};

struct io_epoll {
	struct file			*file;
	int				epfd;
	int				op;
	int				fd;
	struct epoll_event		event;
};

struct io_splice {
	struct file			*file_out;
	struct file			*file_in;
	loff_t				off_out;
	loff_t				off_in;
	u64				len;
	unsigned int			flags;
};

struct io_provide_buf {
	struct file			*file;
	__u64				addr;
	__u32				len;
	__u32				bgid;
	__u16				nbufs;
	__u16				bid;
};

struct io_statx {
	struct file			*file;
	int				dfd;
	unsigned int			mask;
	unsigned int			flags;
	const char __user		*filename;
	struct statx __user		*buffer;
};

struct io_shutdown {
	struct file			*file;
	int				how;
};

struct io_rename {
	struct file			*file;
	int				old_dfd;
	int				new_dfd;
	struct filename			*oldpath;
	struct filename			*newpath;
	int				flags;
};

struct io_unlink {
	struct file			*file;
	int				dfd;
	int				flags;
	struct filename			*filename;
};

struct io_completion {
	struct file			*file;
	u32				cflags;
};

enum {
	REQ_F_FIXED_FILE_BIT	= IOSQE_FIXED_FILE_BIT,
	REQ_F_IO_DRAIN_BIT	= IOSQE_IO_DRAIN_BIT,
	REQ_F_LINK_BIT		= IOSQE_IO_LINK_BIT,
	REQ_F_HARDLINK_BIT	= IOSQE_IO_HARDLINK_BIT,
	REQ_F_FORCE_ASYNC_BIT	= IOSQE_ASYNC_BIT,
	REQ_F_BUFFER_SELECT_BIT	= IOSQE_BUFFER_SELECT_BIT,

	/* first byte is taken by user flags, shift it to not overlap */
	REQ_F_FAIL_BIT		= 8,
	REQ_F_INFLIGHT_BIT,
	REQ_F_CUR_POS_BIT,
	REQ_F_NOWAIT_BIT,
	REQ_F_LINK_TIMEOUT_BIT,
	REQ_F_NEED_CLEANUP_BIT,
	REQ_F_POLLED_BIT,
	REQ_F_BUFFER_SELECTED_BIT,
	REQ_F_LTIMEOUT_ACTIVE_BIT,
	REQ_F_COMPLETE_INLINE_BIT,
	REQ_F_REISSUE_BIT,
	REQ_F_DONT_REISSUE_BIT,
	REQ_F_CREDS_BIT,
	/* keep async read/write and isreg together and in order */
	REQ_F_NOWAIT_READ_BIT,
	REQ_F_NOWAIT_WRITE_BIT,
	REQ_F_ISREG_BIT,

	/* not a real bit, just to check we're not overflowing the space */
	__REQ_F_LAST_BIT,
};

enum {
	/* ctx owns file */
	REQ_F_FIXED_FILE	= BIT(REQ_F_FIXED_FILE_BIT),
	/* drain existing IO first */
	REQ_F_IO_DRAIN		= BIT(REQ_F_IO_DRAIN_BIT),
	/* linked sqes */
	REQ_F_LINK		= BIT(REQ_F_LINK_BIT),
	/* doesn't sever on completion < 0 */
	REQ_F_HARDLINK		= BIT(REQ_F_HARDLINK_BIT),
	/* IOSQE_ASYNC */
	REQ_F_FORCE_ASYNC	= BIT(REQ_F_FORCE_ASYNC_BIT),
	/* IOSQE_BUFFER_SELECT */
	REQ_F_BUFFER_SELECT	= BIT(REQ_F_BUFFER_SELECT_BIT),

	/* fail rest of links */
	REQ_F_FAIL		= BIT(REQ_F_FAIL_BIT),
	/* on inflight list, should be cancelled and waited on exit reliably */
	REQ_F_INFLIGHT		= BIT(REQ_F_INFLIGHT_BIT),
	/* read/write uses file position */
	REQ_F_CUR_POS		= BIT(REQ_F_CUR_POS_BIT),
	/* must not punt to workers */
	REQ_F_NOWAIT		= BIT(REQ_F_NOWAIT_BIT),
	/* has or had linked timeout */
	REQ_F_LINK_TIMEOUT	= BIT(REQ_F_LINK_TIMEOUT_BIT),
	/* needs cleanup */
	REQ_F_NEED_CLEANUP	= BIT(REQ_F_NEED_CLEANUP_BIT),
	/* already went through poll handler */
	REQ_F_POLLED		= BIT(REQ_F_POLLED_BIT),
	/* buffer already selected */
	REQ_F_BUFFER_SELECTED	= BIT(REQ_F_BUFFER_SELECTED_BIT),
	/* linked timeout is active, i.e. prepared by link's head */
	REQ_F_LTIMEOUT_ACTIVE	= BIT(REQ_F_LTIMEOUT_ACTIVE_BIT),
	/* completion is deferred through io_comp_state */
	REQ_F_COMPLETE_INLINE	= BIT(REQ_F_COMPLETE_INLINE_BIT),
	/* caller should reissue async */
	REQ_F_REISSUE		= BIT(REQ_F_REISSUE_BIT),
	/* don't attempt request reissue, see io_rw_reissue() */
	REQ_F_DONT_REISSUE	= BIT(REQ_F_DONT_REISSUE_BIT),
	/* supports async reads */
	REQ_F_NOWAIT_READ	= BIT(REQ_F_NOWAIT_READ_BIT),
	/* supports async writes */
	REQ_F_NOWAIT_WRITE	= BIT(REQ_F_NOWAIT_WRITE_BIT),
	/* regular file */
	REQ_F_ISREG		= BIT(REQ_F_ISREG_BIT),
	/* has creds assigned */
	REQ_F_CREDS		= BIT(REQ_F_CREDS_BIT),
};

struct async_poll {
	struct io_poll_iocb	poll;
	struct io_poll_iocb	*double_poll;
};

typedef void (*io_req_tw_func_t)(struct io_kiocb *req);

struct io_task_work {
	union {
		struct io_wq_work_node	node;
		struct llist_node	fallback_node;
	};
	io_req_tw_func_t		func;
};

enum {
	IORING_RSRC_FILE		= 0,
	IORING_RSRC_BUFFER		= 1,
};

struct io_kiocb {
	union {
		struct file		*file;
		struct io_rw		rw;
		struct io_poll_iocb	poll;
		struct io_poll_update	poll_update;
		struct io_accept	accept;
		struct io_sync		sync;
		struct io_cancel	cancel;
		struct io_timeout	timeout;
		struct io_timeout_rem	timeout_rem;
		struct io_connect	connect;
		struct io_sr_msg	sr_msg;
		struct io_open		open;
		struct io_close		close;
		struct io_rsrc_update	rsrc_update;
		struct io_fadvise	fadvise;
		struct io_madvise	madvise;
		struct io_epoll		epoll;
		struct io_splice	splice;
		struct io_provide_buf	pbuf;
		struct io_statx		statx;
		struct io_shutdown	shutdown;
		struct io_rename	rename;
		struct io_unlink	unlink;
		/* use only after cleaning per-op data, see io_clean_op() */
		struct io_completion	compl;
	};

	/* opcode allocated if it needs to store data for async defer */
	void				*async_data;
	u8				opcode;
	/* polled IO has completed */
	u8				iopoll_completed;

	u16				buf_index;
	u32				result;

	struct io_ring_ctx		*ctx;
	unsigned int			flags;
	atomic_t			refs;
	struct task_struct		*task;
	u64				user_data;

	struct io_kiocb			*link;
	struct percpu_ref		*fixed_rsrc_refs;

	/* used with ctx->iopoll_list with reads/writes */
	struct list_head		inflight_entry;
	struct io_task_work		io_task_work;
	/* for polled requests, i.e. IORING_OP_POLL_ADD and async armed poll */
	struct hlist_node		hash_node;
	struct async_poll		*apoll;
	struct io_wq_work		work;
	const struct cred		*creds;

	/* store used ubuf, so we can prevent reloading */
	struct io_mapped_ubuf		*imu;
};

struct io_tctx_node {
	struct list_head	ctx_node;
	struct task_struct	*task;
	struct io_ring_ctx	*ctx;
};

static bool (*klpe_io_cqring_fill_event)(struct io_ring_ctx *ctx, u64 user_data,
				 long res, unsigned int cflags);
static void klpr_io_put_req(struct io_kiocb *req);
static void klpr_io_put_req_deferred(struct io_kiocb *req, int nr);

static int (*klpe___io_register_rsrc_update)(struct io_ring_ctx *ctx, unsigned type,
				     struct io_uring_rsrc_update2 *up,
				     unsigned nr_args);
static void (*klpe_io_clean_op)(struct io_kiocb *req);

static const struct file_operations (*klpe_io_uring_fops);

static inline void req_set_fail(struct io_kiocb *req)
{
	req->flags |= REQ_F_FAIL;
}

static inline bool io_is_timeout_noseq(struct io_kiocb *req)
{
	return !req->timeout.off;
}

static void (*klpe___io_commit_cqring_flush)(struct io_ring_ctx *ctx);

static inline void klpr_io_commit_cqring(struct io_ring_ctx *ctx)
{
	if (unlikely(ctx->off_timeout_used || ctx->drain_active))
		(*klpe___io_commit_cqring_flush)(ctx);
	/* order cqe stores with ring update */
	smp_store_release(&ctx->rings->cq.tail, ctx->cached_cq_tail);
}

static void (*klpe_io_cqring_ev_posted)(struct io_ring_ctx *ctx);

#define req_ref_zero_or_close_to_overflow(req)	\
	((unsigned int) atomic_read(&(req->refs)) + 127u <= 127u)

static inline bool req_ref_sub_and_test(struct io_kiocb *req, int refs)
{
	WARN_ON_ONCE(req_ref_zero_or_close_to_overflow(req));
	return atomic_sub_and_test(refs, &req->refs);
}

static inline bool req_ref_put_and_test(struct io_kiocb *req)
{
	WARN_ON_ONCE(req_ref_zero_or_close_to_overflow(req));
	return atomic_dec_and_test(&req->refs);
}

static bool (*klpe_io_cqring_fill_event)(struct io_ring_ctx *ctx, u64 user_data,
					  long res, unsigned int cflags);

static void (*klpe_io_req_complete_post)(struct io_kiocb *req, long res,
				 unsigned int cflags);

static inline bool io_req_needs_clean(struct io_kiocb *req)
{
	return req->flags & IO_REQ_CLEAN_FLAGS;
}

static void klpr_io_req_complete_state(struct io_kiocb *req, long res,
				  unsigned int cflags)
{
	if (io_req_needs_clean(req))
		(*klpe_io_clean_op)(req);
	req->result = res;
	req->compl.cflags = cflags;
	req->flags |= REQ_F_COMPLETE_INLINE;
}

static inline void klpr___io_req_complete(struct io_kiocb *req, unsigned issue_flags,
				     long res, unsigned cflags)
{
	if (issue_flags & IO_URING_F_COMPLETE_DEFER)
		klpr_io_req_complete_state(req, res, cflags);
	else
		(*klpe_io_req_complete_post)(req, res, cflags);
}

static inline void klpr_io_req_complete(struct io_kiocb *req, long res)
{
	klpr___io_req_complete(req, 0, res, 0);
}

static inline void io_put_file(struct file *file)
{
	if (file)
		fput(file);
}

static void (*klpe_io_req_task_work_add)(struct io_kiocb *req);

static void (*klpe_io_free_req)(struct io_kiocb *req);

static inline void klpr_io_put_req(struct io_kiocb *req)
{
	if (req_ref_put_and_test(req))
		(*klpe_io_free_req)(req);
}

static inline void klpr_io_put_req_deferred(struct io_kiocb *req, int refs)
{
	if (req_ref_sub_and_test(req, refs)) {
		req->io_task_work.func = (*klpe_io_free_req);
		(*klpe_io_req_task_work_add)(req);
	}
}

static void io_iopoll_req_issued(struct io_kiocb *req)
{
	struct io_ring_ctx *ctx = req->ctx;
	const bool in_async = io_wq_current_is_worker();

	/* workqueue context doesn't hold uring_lock, grab it now */
	if (unlikely(in_async))
		mutex_lock(&ctx->uring_lock);

	/*
	 * Track whether we have multiple files in our lists. This will impact
	 * how we do polling eventually, not spinning if we're on potentially
	 * different devices.
	 */
	if (list_empty(&ctx->iopoll_list)) {
		ctx->poll_multi_queue = false;
	} else if (!ctx->poll_multi_queue) {
		struct io_kiocb *list_req;
		unsigned int queue_num0, queue_num1;

		list_req = list_first_entry(&ctx->iopoll_list, struct io_kiocb,
						inflight_entry);

		if (list_req->file != req->file) {
			ctx->poll_multi_queue = true;
		} else {
			queue_num0 = blk_qc_t_to_queue_num(list_req->rw.kiocb.ki_cookie);
			queue_num1 = blk_qc_t_to_queue_num(req->rw.kiocb.ki_cookie);
			if (queue_num0 != queue_num1)
				ctx->poll_multi_queue = true;
		}
	}

	/*
	 * For fast devices, IO may have already completed. If it has, add
	 * it to the front so we find it first.
	 */
	if (READ_ONCE(req->iopoll_completed))
		list_add(&req->inflight_entry, &ctx->iopoll_list);
	else
		list_add_tail(&req->inflight_entry, &ctx->iopoll_list);

	if (unlikely(in_async)) {
		/*
		 * If IORING_SETUP_SQPOLL is enabled, sqes are either handle
		 * in sq thread task context or in io worker task context. If
		 * current task context is sq thread, we don't need to check
		 * whether should wake up sq thread.
		 */
		if ((ctx->flags & IORING_SETUP_SQPOLL) &&
		    wq_has_sleeper(&ctx->sq_data->wait))
			wake_up(&ctx->sq_data->wait);

		mutex_unlock(&ctx->uring_lock);
	}
}

static void io_ring_submit_unlock(struct io_ring_ctx *ctx, bool needs_lock)
{
	if (needs_lock)
		mutex_unlock(&ctx->uring_lock);
}

static void io_ring_submit_lock(struct io_ring_ctx *ctx, bool needs_lock)
{
	/*
	 * "Normal" inline submissions always hold the uring_lock, since we
	 * grab it from the system call. Same is true for the SQPOLL offload.
	 * The only exception is when we've detached the request and issue it
	 * from an async worker thread, grab the lock for that case.
	 */
	if (needs_lock)
		mutex_lock(&ctx->uring_lock);
}

static int (*klpe_io_read)(struct io_kiocb *req, unsigned int issue_flags);

static int (*klpe_io_write)(struct io_kiocb *req, unsigned int issue_flags);

static int klpr_io_renameat(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_rename *ren = &req->rename;
	int ret;

	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;

	ret = (*klpe_do_renameat2)(ren->old_dfd, ren->oldpath, ren->new_dfd,
				ren->newpath, ren->flags);

	req->flags &= ~REQ_F_NEED_CLEANUP;
	if (ret < 0)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
}

static int klpr_io_unlinkat(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_unlink *un = &req->unlink;
	int ret;

	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;

	if (un->flags & AT_REMOVEDIR)
		ret = (*klpe_do_rmdir)(un->dfd, un->filename);
	else
		ret = (*klpe_do_unlinkat)(un->dfd, un->filename);

	req->flags &= ~REQ_F_NEED_CLEANUP;
	if (ret < 0)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
}

static int klpr_io_shutdown(struct io_kiocb *req, unsigned int issue_flags)
{
#if defined(CONFIG_NET)
	struct socket *sock;
	int ret;

	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;

	sock = sock_from_file(req->file);
	if (unlikely(!sock))
		return -ENOTSOCK;

	ret = (*klpe___sys_shutdown_sock)(sock, req->shutdown.how);
	if (ret < 0)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
#else
#error "klp-ccp: non-taken branch"
#endif
}

static int klpr_io_tee(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_splice *sp = &req->splice;
	struct file *in = sp->file_in;
	struct file *out = sp->file_out;
	unsigned int flags = sp->flags & ~SPLICE_F_FD_IN_FIXED;
	long ret = 0;

	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;
	if (sp->len)
		ret = (*klpe_do_tee)(in, out, sp->len, flags);

	if (!(sp->flags & SPLICE_F_FD_IN_FIXED))
		io_put_file(in);
	req->flags &= ~REQ_F_NEED_CLEANUP;

	if (ret != sp->len)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
}

static int klpr_io_splice(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_splice *sp = &req->splice;
	struct file *in = sp->file_in;
	struct file *out = sp->file_out;
	unsigned int flags = sp->flags & ~SPLICE_F_FD_IN_FIXED;
	loff_t *poff_in, *poff_out;
	long ret = 0;

	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;

	poff_in = (sp->off_in == -1) ? NULL : &sp->off_in;
	poff_out = (sp->off_out == -1) ? NULL : &sp->off_out;

	if (sp->len)
		ret = (*klpe_do_splice)(in, poff_in, out, poff_out, sp->len, flags);

	if (!(sp->flags & SPLICE_F_FD_IN_FIXED))
		io_put_file(in);
	req->flags &= ~REQ_F_NEED_CLEANUP;

	if (ret != sp->len)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
}

static int klpr_io_nop(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_ring_ctx *ctx = req->ctx;

	if (unlikely(ctx->flags & IORING_SETUP_IOPOLL))
		return -EINVAL;

	klpr___io_req_complete(req, issue_flags, 0, 0);
	return 0;
}

static int klpr_io_fsync(struct io_kiocb *req, unsigned int issue_flags)
{
	loff_t end = req->sync.off + req->sync.len;
	int ret;

	/* fsync always requires a blocking context */
	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;

	ret = vfs_fsync_range(req->file, req->sync.off,
				end > 0 ? end : LLONG_MAX,
				req->sync.flags & IORING_FSYNC_DATASYNC);
	if (ret < 0)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
}

static int klpr_io_fallocate(struct io_kiocb *req, unsigned int issue_flags)
{
	int ret;

	/* fallocate always requiring blocking context */
	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;
	ret = vfs_fallocate(req->file, req->sync.mode, req->sync.off,
				req->sync.len);
	if (ret < 0)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
}

static int (*klpe_io_openat2)(struct io_kiocb *req, unsigned int issue_flags);

static int klpr_io_openat(struct io_kiocb *req, unsigned int issue_flags)
{
	return (*klpe_io_openat2)(req, issue_flags);
}

static int __io_remove_buffers(struct io_ring_ctx *ctx, struct io_buffer *buf,
			       int bgid, unsigned nbufs)
{
	unsigned i = 0;

	/* shouldn't happen */
	if (!nbufs)
		return 0;

	/* the head kbuf is the list itself */
	while (!list_empty(&buf->list)) {
		struct io_buffer *nxt;

		nxt = list_first_entry(&buf->list, struct io_buffer, list);
		list_del(&nxt->list);
		kfree(nxt);
		if (++i == nbufs)
			return i;
	}
	i++;
	kfree(buf);
	xa_erase(&ctx->io_buffers, bgid);

	return i;
}

static int klpr_io_remove_buffers(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_provide_buf *p = &req->pbuf;
	struct io_ring_ctx *ctx = req->ctx;
	struct io_buffer *head;
	int ret = 0;
	bool force_nonblock = issue_flags & IO_URING_F_NONBLOCK;

	io_ring_submit_lock(ctx, !force_nonblock);

	lockdep_assert_held(&ctx->uring_lock);

	ret = -ENOENT;
	head = xa_load(&ctx->io_buffers, p->bgid);
	if (head)
		ret = __io_remove_buffers(ctx, head, p->bgid, p->nbufs);
	if (ret < 0)
		req_set_fail(req);

	/* complete before unlock, IOPOLL may need the lock */
	klpr___io_req_complete(req, issue_flags, ret, 0);
	io_ring_submit_unlock(ctx, !force_nonblock);
	return 0;
}

static int io_add_buffers(struct io_provide_buf *pbuf, struct io_buffer **head)
{
	struct io_buffer *buf;
	u64 addr = pbuf->addr;
	int i, bid = pbuf->bid;

	for (i = 0; i < pbuf->nbufs; i++) {
		buf = kmalloc(sizeof(*buf), GFP_KERNEL_ACCOUNT);
		if (!buf)
			break;

		buf->addr = addr;
		buf->len = min_t(__u32, pbuf->len, MAX_RW_COUNT);
		buf->bid = bid;
		addr += pbuf->len;
		bid++;
		if (!*head) {
			INIT_LIST_HEAD(&buf->list);
			*head = buf;
		} else {
			list_add_tail(&buf->list, &(*head)->list);
		}
		cond_resched();
	}

	return i ? i : -ENOMEM;
}

static int klpr_io_provide_buffers(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_provide_buf *p = &req->pbuf;
	struct io_ring_ctx *ctx = req->ctx;
	struct io_buffer *head, *list;
	int ret = 0;
	bool force_nonblock = issue_flags & IO_URING_F_NONBLOCK;

	io_ring_submit_lock(ctx, !force_nonblock);

	lockdep_assert_held(&ctx->uring_lock);

	list = head = xa_load(&ctx->io_buffers, p->bgid);

	ret = io_add_buffers(p, &head);
	if (ret >= 0 && !list) {
		ret = xa_insert(&ctx->io_buffers, p->bgid, head, GFP_KERNEL);
		if (ret < 0)
			__io_remove_buffers(ctx, head, p->bgid, -1U);
	}
	if (ret < 0)
		req_set_fail(req);
	/* complete before unlock, IOPOLL may need the lock */
	klpr___io_req_complete(req, issue_flags, ret, 0);
	io_ring_submit_unlock(ctx, !force_nonblock);
	return 0;
}

static int klpr_io_epoll_ctl(struct io_kiocb *req, unsigned int issue_flags)
{
#if defined(CONFIG_EPOLL)
	struct io_epoll *ie = &req->epoll;
	int ret;
	bool force_nonblock = issue_flags & IO_URING_F_NONBLOCK;

	ret = (*klpe_do_epoll_ctl)(ie->epfd, ie->op, ie->fd, &ie->event, force_nonblock);
	if (force_nonblock && ret == -EAGAIN)
		return -EAGAIN;

	if (ret < 0)
		req_set_fail(req);
	klpr___io_req_complete(req, issue_flags, ret, 0);
	return 0;
#else
#error "klp-ccp: non-taken branch"
#endif
}

static int klpr_io_madvise(struct io_kiocb *req, unsigned int issue_flags)
{
#if defined(CONFIG_ADVISE_SYSCALLS) && defined(CONFIG_MMU)
	struct io_madvise *ma = &req->madvise;
	int ret;

	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;

	ret = (*klpe_do_madvise)(current->mm, ma->addr, ma->len, ma->advice);
	if (ret < 0)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
#else
#error "klp-ccp: non-taken branch"
#endif
}

static int klpr_io_fadvise(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_fadvise *fa = &req->fadvise;
	int ret;

	if (issue_flags & IO_URING_F_NONBLOCK) {
		switch (fa->advice) {
		case POSIX_FADV_NORMAL:
		case POSIX_FADV_RANDOM:
		case POSIX_FADV_SEQUENTIAL:
			break;
		default:
			return -EAGAIN;
		}
	}

	ret = vfs_fadvise(req->file, fa->offset, fa->len, fa->advice);
	if (ret < 0)
		req_set_fail(req);
	klpr___io_req_complete(req, issue_flags, ret, 0);
	return 0;
}

static int klpr_io_statx(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_statx *ctx = &req->statx;
	int ret;

	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;

	ret = (*klpe_do_statx)(ctx->dfd, ctx->filename, ctx->flags, ctx->mask,
		       ctx->buffer);

	if (ret < 0)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
}

static int klpr_io_close(struct io_kiocb *req, unsigned int issue_flags)
{
	struct files_struct *files = current->files;
	struct io_close *close = &req->close;
	struct fdtable *fdt;
	struct file *file = NULL;
	int ret = -EBADF;

	spin_lock(&files->file_lock);
	fdt = files_fdtable(files);
	if (close->fd >= fdt->max_fds) {
		spin_unlock(&files->file_lock);
		goto err;
	}
	file = fdt->fd[close->fd];
	if (!file || file->f_op == &(*klpe_io_uring_fops)) {
		spin_unlock(&files->file_lock);
		file = NULL;
		goto err;
	}

	/* if the file has a flush method, be safe and punt to async */
	if (file->f_op->flush && (issue_flags & IO_URING_F_NONBLOCK)) {
		spin_unlock(&files->file_lock);
		return -EAGAIN;
	}

	ret = (*klpe___close_fd_get_file)(close->fd, &file);
	spin_unlock(&files->file_lock);
	if (ret < 0) {
		if (ret == -ENOENT)
			ret = -EBADF;
		goto err;
	}

	/* No ->flush() or already async, safely close from here */
	ret = filp_close(file, current->files);
err:
	if (ret < 0)
		req_set_fail(req);
	if (file)
		fput(file);
	klpr___io_req_complete(req, issue_flags, ret, 0);
	return 0;
}

static int klpr_io_sync_file_range(struct io_kiocb *req, unsigned int issue_flags)
{
	int ret;

	/* sync_file_range always requires a blocking context */
	if (issue_flags & IO_URING_F_NONBLOCK)
		return -EAGAIN;

	ret = (*klpe_sync_file_range)(req->file, req->sync.off, req->sync.len,
				req->sync.flags);
	if (ret < 0)
		req_set_fail(req);
	klpr_io_req_complete(req, ret);
	return 0;
}

#if defined(CONFIG_NET)

static int (*klpe_io_sendmsg)(struct io_kiocb *req, unsigned int issue_flags);

static int (*klpe_io_send)(struct io_kiocb *req, unsigned int issue_flags);

static int (*klpe_io_recvmsg)(struct io_kiocb *req, unsigned int issue_flags);

static int (*klpe_io_recv)(struct io_kiocb *req, unsigned int issue_flags);

static int klpr_io_accept(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_accept *accept = &req->accept;
	bool force_nonblock = issue_flags & IO_URING_F_NONBLOCK;
	unsigned int file_flags = force_nonblock ? O_NONBLOCK : 0;
	int ret;

	if (req->file->f_flags & O_NONBLOCK)
		req->flags |= REQ_F_NOWAIT;

	ret = (*klpe___sys_accept4_file)(req->file, file_flags, accept->addr,
					accept->addr_len, accept->flags,
					accept->nofile);
	if (ret == -EAGAIN && force_nonblock)
		return -EAGAIN;
	if (ret < 0) {
		if (ret == -ERESTARTSYS)
			ret = -EINTR;
		req_set_fail(req);
	}
	klpr___io_req_complete(req, issue_flags, ret, 0);
	return 0;
}

static int (*klpe_io_connect)(struct io_kiocb *req, unsigned int issue_flags);

#else /* !CONFIG_NET */
#error "klp-ccp: non-taken branch"
#endif /* CONFIG_NET */

struct io_poll_table {
	struct poll_table_struct pt;
	struct io_kiocb *req;
	int nr_entries;
	int error;
};

static struct io_poll_iocb *io_poll_get_single(struct io_kiocb *req)
{
	if (req->opcode == IORING_OP_POLL_ADD)
		return &req->poll;
	return &req->apoll->poll;
}

static void (*klpe_io_poll_remove_double)(struct io_kiocb *req);

static bool (*klpe_io_poll_complete)(struct io_kiocb *req, __poll_t mask);

#define IO_POLL_UNMASK	(EPOLLERR|EPOLLHUP|EPOLLNVAL|EPOLLRDHUP)

static __poll_t (*klpe___io_arm_poll_handler)(struct io_kiocb *req,
				      struct io_poll_iocb *poll,
				      struct io_poll_table *ipt, __poll_t mask,
				      wait_queue_func_t wake_func);

static const struct file_operations (*klpe_signalfd_fops);

static bool klpp___io_poll_remove_one(struct io_kiocb *req,
				 struct io_poll_iocb *poll, bool do_cancel)
	__must_hold(&req->ctx->completion_lock)
{
	bool do_complete = false;

	if (!poll->head)
		return false;

	/*
	 * Fix CVE-2021-39698
	 *  -1 line, +79 lines
	 */
	if (poll->file->f_op != &(*klpe_signalfd_fops)) {
		spin_lock_irq(&poll->head->lock);
	} else {
		/*
		 * In case signalfd_cleanup() has sent out POLLFREE
		 * events already, poll->head might not exist anymore
		 * and taking its ->lock below would cause a UAF, so
		 * some extra care is necessary.
		 *
		 * First note that in the case of signalfd, poll->head
		 * points to some sighand_struct's ->signalfd_wqh and
		 * sighand_struct instances are allocated from a
		 * SLAB_TYPESAFE_BY_RCU slab. Thus, if it can be
		 * proven under a rcu_read_lock() that a pointer to
		 * some ->signalfd_wqh points to a valid instance,
		 * then it will continue to do so for the duration of
		 * the RCU read side critical section. The slab object
		 * might get freed and reassigned in the meanwhile,
		 * but its ->lock may still be taken.
		 *
		 * When we are here and are holding the
		 * ->completion_lock, then this req's ->hash_node is
		 * still on some ctx->cancel_hash[] list, from which
		 * it follows that io_{poll,async}_task_func() hasn't
		 * got to run yet, a task_work or fallback work_struct
		 * might be pending though.
		 *
		 * Assuming that __cleanup_sighand() has begun
		 * execution, i.e. that req->task->sighand == NULL
		 * already, two scenarios are possible:
		 * 1.) signalfd_cleanup() has sent out a POLLFREE
		 *     event for req and the __io_async_wake()
		 *     callback (*) removed the poll entry from the
		 *     wait_queue_head. In this case,
		 *     __cleanup_sighand() might have made it past the
		 *     sighand_struct deallocation and poll->head must
		 *     not get accessed anymore. Note that in this
		 *     case, the pending io_{poll,async}_task_func()
		 *     would see the PF_EXITING and refrain from
		 *     reenqueueing the poll entry. So it's off the
		 *     ->signal_wqh and it will stay that way, it's
		 *     just that poll->head->lock must not be taken.
		 *
		 *     (*) invoked either for the POLLFREE event or
		 *         for some prior, unrelated event
		 *
		 * 2.) signalfd_cleanup() has not yet sent out a
		 *     POLLFREE event for req and the latter's poll
		 *     entry is still on ->signalfd_wqh. The
		 *     ->signalfd_wqh might go away anytime soon, but
		 *     its backing memory is guaranteed to hold *some*
		 *     wait_queue_head for the duration of the RCU
		 *     read side critical section below.
		 */
		rcu_read_lock();
		if (req->task && !READ_ONCE(req->task->sighand)) {
			if (list_empty(&poll->wait.entry)) {
				/* Case 1.) from above. */
				rcu_read_unlock();

				/*
				 * Emulate original behavior, but skip
				 * the poll->wait.entry manipulation.
				 */
				if (do_cancel)
					WRITE_ONCE(poll->canceled, true);
				hash_del(&req->hash_node);
				return false;
			}
		}

		/*
		 * Possibly case 2.) from above, poll->head still
		 * points to *some* valid sighand_struct's
		 * ->signalfd_wqh, not necessarily the original
		 * one. Its ->lock may be taken either way.
		 */
		spin_lock(&poll->head->lock);
	}

	if (do_cancel)
		WRITE_ONCE(poll->canceled, true);
	if (!list_empty(&poll->wait.entry)) {
		list_del_init(&poll->wait.entry);
		do_complete = true;
	}
	spin_unlock_irq(&poll->head->lock);
	/*
	 * Fix CVE-2021-39698
	 *  +2 lines
	 */
	if (poll->file->f_op == &(*klpe_signalfd_fops))
		rcu_read_unlock();

	hash_del(&req->hash_node);
	return do_complete;
}

bool klpp_io_poll_remove_one(struct io_kiocb *req)
	__must_hold(&req->ctx->completion_lock)
{
	int refs;
	bool do_complete;

	(*klpe_io_poll_remove_double)(req);
	do_complete = klpp___io_poll_remove_one(req, io_poll_get_single(req), true);

	if (do_complete) {
		(*klpe_io_cqring_fill_event)(req->ctx, req->user_data, -ECANCELED, 0);
		klpr_io_commit_cqring(req->ctx);
		req_set_fail(req);

		/* non-poll requests have submit ref still */
		refs = 1 + (req->opcode != IORING_OP_POLL_ADD);
		klpr_io_put_req_deferred(req, refs);
	}
	return do_complete;
}

static struct io_kiocb *io_poll_find(struct io_ring_ctx *ctx, __u64 sqe_addr,
				     bool poll_only)
	__must_hold(&ctx->completion_lock)
{
	struct hlist_head *list;
	struct io_kiocb *req;

	list = &ctx->cancel_hash[hash_long(sqe_addr, ctx->cancel_hash_bits)];
	hlist_for_each_entry(req, list, hash_node) {
		if (sqe_addr != req->user_data)
			continue;
		if (poll_only && req->opcode != IORING_OP_POLL_ADD)
			continue;
		return req;
	}
	return NULL;
}

static int klpr_io_poll_cancel(struct io_ring_ctx *ctx, __u64 sqe_addr,
			  bool poll_only)
	__must_hold(&ctx->completion_lock)
{
	struct io_kiocb *req;

	req = io_poll_find(ctx, sqe_addr, poll_only);
	if (!req)
		return -ENOENT;
	if (klpp_io_poll_remove_one(req))
		return 0;

	return -EALREADY;
}

static int (*klpe_io_poll_wake)(struct wait_queue_entry *wait, unsigned mode, int sync,
			void *key);

static void (*klpe_io_poll_queue_proc)(struct file *file, struct wait_queue_head *head,
			       struct poll_table_struct *p);

static int klpr_io_poll_add(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_poll_iocb *poll = &req->poll;
	struct io_ring_ctx *ctx = req->ctx;
	struct io_poll_table ipt;
	__poll_t mask;

	ipt.pt._qproc = (*klpe_io_poll_queue_proc);

	mask = (*klpe___io_arm_poll_handler)(req, &req->poll, &ipt, poll->events,
					(*klpe_io_poll_wake));

	if (mask) { /* no async, we'd stolen it */
		ipt.error = 0;
		(*klpe_io_poll_complete)(req, mask);
	}
	spin_unlock(&ctx->completion_lock);

	if (mask) {
		(*klpe_io_cqring_ev_posted)(ctx);
		if (poll->events & EPOLLONESHOT)
			klpr_io_put_req(req);
	}
	return ipt.error;
}

static int klpp_io_poll_update(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_ring_ctx *ctx = req->ctx;
	struct io_kiocb *preq;
	bool completing;
	int ret;

	spin_lock(&ctx->completion_lock);
	preq = io_poll_find(ctx, req->poll_update.old_user_data, true);
	if (!preq) {
		ret = -ENOENT;
		goto err;
	}

	if (!req->poll_update.update_events && !req->poll_update.update_user_data) {
		completing = true;
		ret = klpp_io_poll_remove_one(preq) ? 0 : -EALREADY;
		goto err;
	}

	/*
	 * Don't allow racy completion with singleshot, as we cannot safely
	 * update those. For multishot, if we're racing with completion, just
	 * let completion re-add it.
	 */
	completing = !klpp___io_poll_remove_one(preq, &preq->poll, false);
	if (completing && (preq->poll.events & EPOLLONESHOT)) {
		ret = -EALREADY;
		goto err;
	}
	/* we now have a detached poll request. reissue. */
	ret = 0;
err:
	if (ret < 0) {
		spin_unlock(&ctx->completion_lock);
		req_set_fail(req);
		klpr_io_req_complete(req, ret);
		return 0;
	}
	/* only mask one event flags, keep behavior flags */
	if (req->poll_update.update_events) {
		preq->poll.events &= ~0xffff;
		preq->poll.events |= req->poll_update.events & 0xffff;
		preq->poll.events |= IO_POLL_UNMASK;
	}
	if (req->poll_update.update_user_data)
		preq->user_data = req->poll_update.new_user_data;
	spin_unlock(&ctx->completion_lock);

	/* complete update request, we're done with it */
	klpr_io_req_complete(req, ret);

	if (!completing) {
		ret = klpr_io_poll_add(preq, issue_flags);
		if (ret < 0) {
			req_set_fail(preq);
			klpr_io_req_complete(preq, ret);
		}
	}
	return 0;
}

static enum hrtimer_restart (*klpe_io_timeout_fn)(struct hrtimer *timer);

static struct io_kiocb *(*klpe_io_timeout_extract)(struct io_ring_ctx *ctx,
					   __u64 user_data);

static int (*klpe_io_timeout_cancel)(struct io_ring_ctx *ctx, __u64 user_data);

static int klpr_io_timeout_update(struct io_ring_ctx *ctx, __u64 user_data,
			     struct timespec64 *ts, enum hrtimer_mode mode)
	__must_hold(&ctx->timeout_lock)
{
	struct io_kiocb *req = (*klpe_io_timeout_extract)(ctx, user_data);
	struct io_timeout_data *data;

	if (IS_ERR(req))
		return PTR_ERR(req);

	req->timeout.off = 0; /* noseq */
	data = req->async_data;
	list_add_tail(&req->timeout.list, &ctx->timeout_list);
	hrtimer_init(&data->timer, CLOCK_MONOTONIC, mode);
	data->timer.function = (*klpe_io_timeout_fn);
	hrtimer_start(&data->timer, timespec64_to_ktime(*ts), mode);
	return 0;
}

static inline enum hrtimer_mode io_translate_timeout_mode(unsigned int flags)
{
	return (flags & IORING_TIMEOUT_ABS) ? HRTIMER_MODE_ABS
					    : HRTIMER_MODE_REL;
}

static int klpr_io_timeout_remove(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_timeout_rem *tr = &req->timeout_rem;
	struct io_ring_ctx *ctx = req->ctx;
	int ret;

	spin_lock_irq(&ctx->timeout_lock);
	if (!(req->timeout_rem.flags & IORING_TIMEOUT_UPDATE))
		ret = (*klpe_io_timeout_cancel)(ctx, tr->addr);
	else
		ret = klpr_io_timeout_update(ctx, tr->addr, &tr->ts,
					io_translate_timeout_mode(tr->flags));
	spin_unlock_irq(&ctx->timeout_lock);

	spin_lock(&ctx->completion_lock);
	(*klpe_io_cqring_fill_event)(ctx, req->user_data, ret, 0);
	klpr_io_commit_cqring(ctx);
	spin_unlock(&ctx->completion_lock);
	(*klpe_io_cqring_ev_posted)(ctx);
	if (ret < 0)
		req_set_fail(req);
	klpr_io_put_req(req);
	return 0;
}

static int klpr_io_timeout(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_ring_ctx *ctx = req->ctx;
	struct io_timeout_data *data = req->async_data;
	struct list_head *entry;
	u32 tail, off = req->timeout.off;

	spin_lock_irq(&ctx->timeout_lock);

	/*
	 * sqe->off holds how many events that need to occur for this
	 * timeout event to be satisfied. If it isn't set, then this is
	 * a pure timeout request, sequence isn't used.
	 */
	if (io_is_timeout_noseq(req)) {
		entry = ctx->timeout_list.prev;
		goto add;
	}

	tail = ctx->cached_cq_tail - atomic_read(&ctx->cq_timeouts);
	req->timeout.target_seq = tail + off;

	/* Update the last seq here in case io_flush_timeouts() hasn't.
	 * This is safe because ->completion_lock is held, and submissions
	 * and completions are never mixed in the same ->completion_lock section.
	 */
	ctx->cq_last_tm_flush = tail;

	/*
	 * Insertion sort, ensuring the first entry in the list is always
	 * the one we need first.
	 */
	list_for_each_prev(entry, &ctx->timeout_list) {
		struct io_kiocb *nxt = list_entry(entry, struct io_kiocb,
						  timeout.list);

		if (io_is_timeout_noseq(nxt))
			continue;
		/* nxt.seq is behind @tail, otherwise would've been completed */
		if (off >= nxt->timeout.target_seq - tail)
			break;
	}
add:
	list_add(&req->timeout.list, entry);
	data->timer.function = (*klpe_io_timeout_fn);
	hrtimer_start(&data->timer, timespec64_to_ktime(data->ts), data->mode);
	spin_unlock_irq(&ctx->timeout_lock);
	return 0;
}

static int (*klpe_io_async_cancel_one)(struct io_uring_task *tctx, u64 user_data,
			       struct io_ring_ctx *ctx);

static int klpr_io_async_cancel(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_ring_ctx *ctx = req->ctx;
	u64 sqe_addr = req->cancel.addr;
	struct io_tctx_node *node;
	int ret;

	/* tasks should wait for their io-wq threads, so safe w/o sync */
	ret = (*klpe_io_async_cancel_one)(req->task->io_uring, sqe_addr, ctx);
	spin_lock(&ctx->completion_lock);
	if (ret != -ENOENT)
		goto done;
	spin_lock_irq(&ctx->timeout_lock);
	ret = (*klpe_io_timeout_cancel)(ctx, sqe_addr);
	spin_unlock_irq(&ctx->timeout_lock);
	if (ret != -ENOENT)
		goto done;
	ret = klpr_io_poll_cancel(ctx, sqe_addr, false);
	if (ret != -ENOENT)
		goto done;
	spin_unlock(&ctx->completion_lock);

	/* slow path, try all io-wq's */
	io_ring_submit_lock(ctx, !(issue_flags & IO_URING_F_NONBLOCK));
	ret = -ENOENT;
	list_for_each_entry(node, &ctx->tctx_list, ctx_node) {
		struct io_uring_task *tctx = node->task->io_uring;

		ret = (*klpe_io_async_cancel_one)(tctx, req->cancel.addr, ctx);
		if (ret != -ENOENT)
			break;
	}
	io_ring_submit_unlock(ctx, !(issue_flags & IO_URING_F_NONBLOCK));

	spin_lock(&ctx->completion_lock);
done:
	(*klpe_io_cqring_fill_event)(ctx, req->user_data, ret, 0);
	klpr_io_commit_cqring(ctx);
	spin_unlock(&ctx->completion_lock);
	(*klpe_io_cqring_ev_posted)(ctx);

	if (ret < 0)
		req_set_fail(req);
	klpr_io_put_req(req);
	return 0;
}

static int klpr_io_files_update(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_ring_ctx *ctx = req->ctx;
	struct io_uring_rsrc_update2 up;
	int ret;

	up.offset = req->rsrc_update.offset;
	up.data = req->rsrc_update.arg;
	up.nr = 0;
	up.tags = 0;
	up.resv = 0;

	io_ring_submit_lock(ctx, !(issue_flags & IO_URING_F_NONBLOCK));
	ret = (*klpe___io_register_rsrc_update)(ctx, IORING_RSRC_FILE,
					&up, req->rsrc_update.nr_args);
	io_ring_submit_unlock(ctx, !(issue_flags & IO_URING_F_NONBLOCK));

	if (ret < 0)
		req_set_fail(req);
	klpr___io_req_complete(req, issue_flags, ret, 0);
	return 0;
}

static void (*klpe_io_clean_op)(struct io_kiocb *req);

int klpp_io_issue_sqe(struct io_kiocb *req, unsigned int issue_flags)
{
	struct io_ring_ctx *ctx = req->ctx;
	const struct cred *creds = NULL;
	int ret;

	if ((req->flags & REQ_F_CREDS) && req->creds != current_cred())
		creds = override_creds(req->creds);

	switch (req->opcode) {
	case IORING_OP_NOP:
		ret = klpr_io_nop(req, issue_flags);
		break;
	case IORING_OP_READV:
	case IORING_OP_READ_FIXED:
	case IORING_OP_READ:
		ret = (*klpe_io_read)(req, issue_flags);
		break;
	case IORING_OP_WRITEV:
	case IORING_OP_WRITE_FIXED:
	case IORING_OP_WRITE:
		ret = (*klpe_io_write)(req, issue_flags);
		break;
	case IORING_OP_FSYNC:
		ret = klpr_io_fsync(req, issue_flags);
		break;
	case IORING_OP_POLL_ADD:
		ret = klpr_io_poll_add(req, issue_flags);
		break;
	case IORING_OP_POLL_REMOVE:
		ret = klpp_io_poll_update(req, issue_flags);
		break;
	case IORING_OP_SYNC_FILE_RANGE:
		ret = klpr_io_sync_file_range(req, issue_flags);
		break;
	case IORING_OP_SENDMSG:
		ret = (*klpe_io_sendmsg)(req, issue_flags);
		break;
	case IORING_OP_SEND:
		ret = (*klpe_io_send)(req, issue_flags);
		break;
	case IORING_OP_RECVMSG:
		ret = (*klpe_io_recvmsg)(req, issue_flags);
		break;
	case IORING_OP_RECV:
		ret = (*klpe_io_recv)(req, issue_flags);
		break;
	case IORING_OP_TIMEOUT:
		ret = klpr_io_timeout(req, issue_flags);
		break;
	case IORING_OP_TIMEOUT_REMOVE:
		ret = klpr_io_timeout_remove(req, issue_flags);
		break;
	case IORING_OP_ACCEPT:
		ret = klpr_io_accept(req, issue_flags);
		break;
	case IORING_OP_CONNECT:
		ret = (*klpe_io_connect)(req, issue_flags);
		break;
	case IORING_OP_ASYNC_CANCEL:
		ret = klpr_io_async_cancel(req, issue_flags);
		break;
	case IORING_OP_FALLOCATE:
		ret = klpr_io_fallocate(req, issue_flags);
		break;
	case IORING_OP_OPENAT:
		ret = klpr_io_openat(req, issue_flags);
		break;
	case IORING_OP_CLOSE:
		ret = klpr_io_close(req, issue_flags);
		break;
	case IORING_OP_FILES_UPDATE:
		ret = klpr_io_files_update(req, issue_flags);
		break;
	case IORING_OP_STATX:
		ret = klpr_io_statx(req, issue_flags);
		break;
	case IORING_OP_FADVISE:
		ret = klpr_io_fadvise(req, issue_flags);
		break;
	case IORING_OP_MADVISE:
		ret = klpr_io_madvise(req, issue_flags);
		break;
	case IORING_OP_OPENAT2:
		ret = (*klpe_io_openat2)(req, issue_flags);
		break;
	case IORING_OP_EPOLL_CTL:
		ret = klpr_io_epoll_ctl(req, issue_flags);
		break;
	case IORING_OP_SPLICE:
		ret = klpr_io_splice(req, issue_flags);
		break;
	case IORING_OP_PROVIDE_BUFFERS:
		ret = klpr_io_provide_buffers(req, issue_flags);
		break;
	case IORING_OP_REMOVE_BUFFERS:
		ret = klpr_io_remove_buffers(req, issue_flags);
		break;
	case IORING_OP_TEE:
		ret = klpr_io_tee(req, issue_flags);
		break;
	case IORING_OP_SHUTDOWN:
		ret = klpr_io_shutdown(req, issue_flags);
		break;
	case IORING_OP_RENAMEAT:
		ret = klpr_io_renameat(req, issue_flags);
		break;
	case IORING_OP_UNLINKAT:
		ret = klpr_io_unlinkat(req, issue_flags);
		break;
	default:
		ret = -EINVAL;
		break;
	}

	if (creds)
		revert_creds(creds);
	if (ret)
		return ret;
	/* If the op doesn't have a file, we're not polling for it */
	if ((ctx->flags & IORING_SETUP_IOPOLL) && req->file)
		io_iopoll_req_issued(req);

	return 0;
}



#include <linux/kernel.h>
#include <linux/module.h>
#include "livepatch_bsc1196959.h"
#include "../kallsyms_relocs.h"

static struct klp_kallsyms_reloc klp_funcs[] = {
	{ "__close_fd_get_file", (void *)&klpe___close_fd_get_file },
	{ "__io_arm_poll_handler", (void *)&klpe___io_arm_poll_handler },
	{ "__io_commit_cqring_flush", (void *)&klpe___io_commit_cqring_flush },
	{ "__io_register_rsrc_update",
	  (void *)&klpe___io_register_rsrc_update },
	{ "__sys_accept4_file", (void *)&klpe___sys_accept4_file },
	{ "__sys_shutdown_sock", (void *)&klpe___sys_shutdown_sock },
	{ "do_epoll_ctl", (void *)&klpe_do_epoll_ctl },
	{ "do_madvise", (void *)&klpe_do_madvise },
	{ "do_renameat2", (void *)&klpe_do_renameat2 },
	{ "do_rmdir", (void *)&klpe_do_rmdir },
	{ "do_splice", (void *)&klpe_do_splice },
	{ "do_statx", (void *)&klpe_do_statx },
	{ "do_tee", (void *)&klpe_do_tee },
	{ "do_unlinkat", (void *)&klpe_do_unlinkat },
	{ "io_async_cancel_one", (void *)&klpe_io_async_cancel_one },
	{ "io_clean_op", (void *)&klpe_io_clean_op },
	{ "io_connect", (void *)&klpe_io_connect },
	{ "io_cqring_ev_posted", (void *)&klpe_io_cqring_ev_posted },
	{ "io_cqring_fill_event", (void *)&klpe_io_cqring_fill_event },
	{ "io_free_req", (void *)&klpe_io_free_req },
	{ "io_openat2", (void *)&klpe_io_openat2 },
	{ "io_poll_complete", (void *)&klpe_io_poll_complete },
	{ "io_poll_queue_proc", (void *)&klpe_io_poll_queue_proc },
	{ "io_poll_remove_double", (void *)&klpe_io_poll_remove_double },
	{ "io_poll_wake", (void *)&klpe_io_poll_wake },
	{ "io_read", (void *)&klpe_io_read },
	{ "io_recv", (void *)&klpe_io_recv },
	{ "io_recvmsg", (void *)&klpe_io_recvmsg },
	{ "io_req_complete_post", (void *)&klpe_io_req_complete_post },
	{ "io_req_task_work_add", (void *)&klpe_io_req_task_work_add },
	{ "io_send", (void *)&klpe_io_send },
	{ "io_sendmsg", (void *)&klpe_io_sendmsg },
	{ "io_timeout_cancel", (void *)&klpe_io_timeout_cancel },
	{ "io_timeout_extract", (void *)&klpe_io_timeout_extract },
	{ "io_timeout_fn", (void *)&klpe_io_timeout_fn },
	{ "io_uring_fops", (void *)&klpe_io_uring_fops },
	{ "io_write", (void *)&klpe_io_write },
	{ "signalfd_fops", (void *)&klpe_signalfd_fops },
	{ "sync_file_range", (void *)&klpe_sync_file_range },
};

int livepatch_bsc1196959_init(void)
{
	return klp_resolve_kallsyms_relocs(klp_funcs, ARRAY_SIZE(klp_funcs));
}
